/**
Copyright 2016 The Chromium Authors. All rights reserved.
Use of this source code is governed by a BSD-style license that can be
found in the LICENSE file.
**/

require("../base/iteration_helpers.js");
require("../base/range.js");
require("../base/running_statistics.js");
require("../base/sorted_array_utils.js");
require("../base/statistics.js");
require("./unit.js");

'use strict';

global.tr.exportTo('tr.v', function() {
  var Range = tr.b.Range;

  var MAX_SOURCE_INFOS = 16;

  function NumericBase(unit) {
    if (!(unit instanceof tr.v.Unit))
      throw new Error('Expected provided unit to be instance of Unit');

    this.unit = unit;
  }

  NumericBase.prototype = {
    asDict: function() {
      var d = {
        unit: this.unit.asJSON()
      };

      this.asDictInto_(d);
      return d;
    }
  };

  NumericBase.fromDict = function(d) {
    if (d.type === 'scalar')
      return ScalarNumeric.fromDict(d);

    throw new Error('Not implemented');
  };

  function NumericBin(parentNumeric, opt_range) {
    this.parentNumeric = parentNumeric;
    this.range = opt_range || (new tr.b.Range());
    this.count = 0;
    this.sourceInfos = [];
  }

  NumericBin.fromDict = function(parentNumeric, d) {
    var n = new NumericBin(parentNumeric);
    n.range.min = d.min;
    n.range.max = d.max;
    n.count = d.count;
    n.sourceInfos = d.sourceInfos;
    return n;
  };

  NumericBin.prototype = {
    add: function(value, sourceInfo) {
      this.count += 1;
      tr.b.Statistics.uniformlySampleStream(this.sourceInfos, this.count,
          sourceInfo, MAX_SOURCE_INFOS);
    },

    addBin: function(other) {
      if (!this.range.equals(other.range))
        throw new Error('Merging incompatible Numeric bins.');
      tr.b.Statistics.mergeSampledStreams(this.sourceInfos, this.count,
          other.sourceInfos, other.count, MAX_SOURCE_INFOS);
      this.count += other.count;
    },

    asDict: function() {
      return {
        min: this.range.min,
        max: this.range.max,
        count: this.count,
        sourceInfos: this.sourceInfos.slice(0)
      };
    },

    asJSON: function() {
      return this.asDict();
    }
  };

  function Numeric(unit, range, binInfo) {
    NumericBase.call(this, unit);

    this.range = range;

    this.numNans = 0;
    this.nanSourceInfos = [];

    this.running = new tr.b.RunningStatistics();
    this.maxCount_ = 0;

    this.underflowBin = binInfo.underflowBin;
    this.centralBins = binInfo.centralBins;
    this.overflowBin = binInfo.overflowBin;

    this.allBins = [];
    this.allBins.push(this.underflowBin);
    this.allBins.push.apply(this.allBins, this.centralBins);
    this.allBins.push(this.overflowBin);

    this.allBins.forEach(function(bin) {
      if (bin.count > this.maxCount_)
        this.maxCount_ = bin.count;
    }, this);

    this.summaryOptions = this.defaultSummaryOptions();
  }

  Numeric.fromDict = function(d) {
    var range = Range.fromExplicitRange(d.min, d.max);
    var binInfo = {};
    binInfo.underflowBin = NumericBin.fromDict(undefined, d.underflowBin);
    binInfo.centralBins = d.centralBins.map(function(binAsDict) {
      return NumericBin.fromDict(undefined, binAsDict);
    });
    binInfo.overflowBin = NumericBin.fromDict(undefined, d.overflowBin);
    var n = new Numeric(tr.v.Unit.fromJSON(d.unit), range, binInfo);
    n.allBins.forEach(function(bin) {
      bin.parentNumeric = n;
    });
    if (d.running)
      n.running = tr.b.RunningStatistics.fromDict(d.running);
    if (d.summaryOptions)
      n.customizeSummaryOptions(d.summaryOptions);
    n.numNans = d.numNans;
    n.nanSourceInfos = d.nanSourceInfos;
    return n;
  };

  Numeric.prototype = {
    __proto__: NumericBase.prototype,

    get numValues() {
      return tr.b.Statistics.sum(this.allBins, function(e) {
        return e.count;
      });
    },

    get average() {
      return this.running.mean;
    },

    get sum() {
      return this.running.sum;
    },

    get maxCount() {
      return this.maxCount_;
    },

    /*
     * Compute an approximation of percentile based on the counts in the bins.
     * If the real percentile lies within |this.range| then the result of
     * the function will deviate from the real percentile by at most
     * the maximum width of the bin(s) within which the point(s)
     * from which the real percentile would be calculated lie.
     * If the real percentile is outside |this.range| then the function
     * returns the closest range limit: |this.range.min| or |this.range.max|.
     *
     * @param {number} percent The percent must be between 0.0 and 1.0.
     */
    getApproximatePercentile: function(percent) {
      if (!(percent >= 0 && percent <= 1))
        throw new Error('percent must be [0,1]');
      if (this.numValues == 0)
        return 0;
      var valuesToSkip = Math.floor((this.numValues - 1) * percent);
      for (var i = 0; i < this.allBins.length; i++) {
        var bin = this.allBins[i];
        valuesToSkip -= bin.count;
        if (valuesToSkip < 0) {
          if (bin === this.underflowBin)
            return bin.range.max;
          else if (bin === this.overflowBin)
            return bin.range.min;
          else
            return bin.range.center;
        }
      }
      throw new Error('Unreachable');
    },

    getInterpolatedCountAt: function(value) {
      var bin = this.getBinForValue(value);
      var idx = this.centralBins.indexOf(bin);
      if (idx < 0) {
        // |value| is in either the underflowBin or the overflowBin.
        // We can't interpolate between infinities.
        return bin.count;
      }

      // |value| must fall between the centers of two bins.
      // The bin whose center is less than |value| will be this:
      var lesserBin = bin;

      // The bin whose center is greater than |value| will be this:
      var greaterBin = bin;

      // One of those bins could be an under/overflow bin.
      // Avoid dealing with Infinities by arbitrarily saying that center of the
      // underflow bin is its range.max, and the center of the overflow bin is
      // its range.min.
      // The centers of bins in |this.centralBins| will default to their
      // |range.center|.

      var lesserBinCenter = undefined;
      var greaterBinCenter = undefined;

      if (value < greaterBin.range.center) {
        if (idx > 0) {
          lesserBin = this.centralBins[idx - 1];
        } else {
          lesserBin = this.underflowBin;
          lesserBinCenter = lesserBin.range.max;
        }
      } else {
        if (idx < (this.centralBins.length - 1)) {
          greaterBin = this.centralBins[idx + 1];
        } else {
          greaterBin = this.overflowBin;
          greaterBinCenter = greaterBin.range.min;
        }
      }

      if (greaterBinCenter === undefined)
        greaterBinCenter = greaterBin.range.center;

      if (lesserBinCenter === undefined)
        lesserBinCenter = lesserBin.range.center;

      value = tr.b.normalize(value, lesserBinCenter, greaterBinCenter);

      return tr.b.lerp(value, lesserBin.count, greaterBin.count);
    },

    getBinForValue: function(value) {
      // Don't use subtraction to avoid arithmetic overflow.
      var binIndex = tr.b.findHighIndexInSortedArray(
          this.allBins, b => value < b.range.max ? -1 : 1);
      return this.allBins[binIndex] || this.overflowBin;
    },

    add: function(value, sourceInfo) {
      if (typeof(value) !== 'number' || isNaN(value)) {
        this.numNans++;
        tr.b.Statistics.uniformlySampleStream(this.nanSourceInfos, this.numNans,
            sourceInfo, MAX_SOURCE_INFOS);
        return;
      }

      var bin = this.getBinForValue(value);
      bin.add(value, sourceInfo);
      this.running.add(value);
      if (bin.count > this.maxCount_)
        this.maxCount_ = bin.count;
    },

    addNumeric: function(other) {
      if (!this.range.equals(other.range) ||
          !this.unit === other.unit ||
          this.allBins.length !== other.allBins.length) {
        throw new Error('Merging incompatible Numerics.');
      }
      tr.b.Statistics.mergeSampledStreams(this.nanSourceInfos, this.numNans,
          other.nanSourceInfos, other.numNans, MAX_SOURCE_INFOS);
      this.numNans += other.numNans;
      this.running = this.running.merge(other.running);
      for (var i = 0; i < this.allBins.length; ++i) {
        this.allBins[i].addBin(other.allBins[i]);
      }
    },

    /**
     * Controls which statistics are exported to dashboard for this numeric.
     * The |summaryOptions| parameter is a dictionary with optional boolean
     * fields |count|, |sum|, |avg|, |std|, |min|, |max| and an optional
     * array field |percentile|.
     * Each percentile should be a number between 0.0 and 1.0.
     * The options not included in the |summaryOptions| will not change.
     */
    customizeSummaryOptions: function(summaryOptions) {
      tr.b.iterItems(summaryOptions, function(key, value) {
        this.summaryOptions[key] = value;
      }, this);
    },

    defaultSummaryOptions: function() {
      return {
        count: true,
        sum: true,
        avg: true,
        std: true,
        min: true,
        max: true,
        percentile: []
      };
    },

    /**
     * Returns an array of {name: string, scalar: ScalarNumeric} values.
     * Each enabled summary option produces the corresponding value:
     * min, max, count, sum, avg, or std.
     * Each percentile 0.x produces pct_0x0.
     * Each percentile 0.xx produces pct_0xx.
     * Each percentile 0.xxy produces pct_0xx_y.
     * Percentile 1.0 produces pct_100.
     */
    getSummarizedScalarNumericsWithNames: function() {
      function statNameToKey(stat) {
        switch (stat) {
          case 'std':
            return 'stddev';
          case 'avg':
            return 'mean';
        }
        return stat;
      }
      /**
       * Converts the given percent to a string in the format specified above.
       * @param {number} percent The percent must be between 0.0 and 1.0.
       */
      function percentToString(percent) {
        if (percent < 0 || percent > 1)
          throw new Error('Percent must be between 0.0 and 1.0');
        switch (percent) {
          case 0:
            return '000';
          case 1:
            return '100';
        }
        var str = percent.toString();
        if (str[1] !== '.')
          throw new Error('Unexpected percent');
        // Pad short strings with zeros.
        str = str + '0'.repeat(Math.max(4 - str.length, 0));
        if (str.length > 4)
          str = str.slice(0, 4) + '_' + str.slice(4);
        return '0' + str.slice(2);
      }

      var results = [];
      tr.b.iterItems(this.summaryOptions, function(stat, option) {
        if (!option)
          return;
        if (stat === 'percentile') {
          option.forEach(function(percent) {
            var percentile = this.getApproximatePercentile(percent);
            results.push({
                name: 'pct_' + percentToString(percent),
                scalar: new tr.v.ScalarNumeric(this.unit, percentile)
            });
          }, this);
        } else {
          var statUnit = stat === 'count' ?
              tr.v.Unit.byName.unitlessNumber_smallerIsBetter : this.unit;
          var key = statNameToKey(stat);
          var statValue = this.running[key];
          if (statValue !== undefined) {
            results.push({
                name: stat,
                scalar: new tr.v.ScalarNumeric(statUnit, statValue)
            });
          }
        }
      }, this);
      return results;
    },

    clone: function() {
      return Numeric.fromDict(this.asDict());
    },

    asDict: function() {
      var d = {
        unit: this.unit.asJSON(),
        type: 'numeric',

        min: this.range.min,
        max: this.range.max,

        numNans: this.numNans,
        nanSourceInfos: this.nanSourceInfos,

        running: this.running.asDict(),
        summaryOptions: this.summaryOptions,

        underflowBin: this.underflowBin.asDict(),
        centralBins: this.centralBins.map(function(bin) {
          return bin.asDict();
        }),
        overflowBin: this.overflowBin.asDict()
      };
      return d;
    },

    asJSON: function() {
      return this.asDict();
    }
  };

  /**
   * Reusable builder for tr.v.Numeric objects.
   *
   * The bins of the numeric are specified by adding the desired boundaries
   * between bins. Initially, the builder has only a single boundary:
   *
   *       minBinBoundary=maxBinBoundary
   *                     |
   *                     |
   *   -MAX_INT <--------|------------------------------------------> +MAX_INT
   *       :  resulting  :                   resulting                    :
   *       :  underflow  :                    overflow                    :
   *       :     bin     :                      bin                       :
   *
   * More boundaries can be added (in increasing order) using addBinBoundary,
   * addLinearBins and addExponentialBins:
   *
   *              minBinBoundary                      maxBinBoundary
   *                     |         |         |     |         |
   *                     |         |         |     |         |
   *   -MAX_INT <--------|---------|---------|-----|---------|------> +MAX_INT
   *       :  resulting  : result. : result. :     : result. : resulting  :
   *       :  underflow  : central : central : ... : central :  overflow  :
   *       :     bin     :  bin 0  :  bin 1  :     : bin N-1 :    bin     :
   *
   * An important feature of the builder is that it's reusable, i.e. it can be
   * used to build multiple numerics with the same unit and bin structure.
   *
   * @constructor
   * @param {!tr.v.Unit} unit Unit of the resulting Numeric(s).
   * @param {number} minBinBoundary The minimum boundary between bins, namely
   *     the underflow bin and the first central bin (or the overflow bin if
   *     no other boundaries are added later).
   */
  function NumericBuilder(unit, minBinBoundary) {
    this.unit_ = unit;
    this.boundaries_ = [minBinBoundary];
  }

  NumericBuilder.prototype = {
    get minBinBoundary() {
      return this.boundaries_[0];
    },

    get maxBinBoundary() {
      return this.boundaries_[this.boundaries_.length - 1];
    },

    /**
     * Add a bin boundary |nextMaxBinBoundary| to the builder.
     *
     * This operation effectively corresponds to appending a new central bin
     * with the range [this.maxBinBoundary*, nextMaxBinBoundary].
     *
     * @param {number} nextMaxBinBoundary The added bin boundary (must be
     *     greater than |this.maxMinBoundary|).
     */
    addBinBoundary: function(nextMaxBinBoundary) {
      if (nextMaxBinBoundary <= this.maxBinBoundary) {
        throw new Error('The added max bin boundary must be larger than ' +
            'the current max boundary');
      }
      this.boundaries_.push(nextMaxBinBoundary);

      return this;
    },

    /**
     * Add |binCount| linearly scaled bin boundaries up to |nextMaxBinBoundary|
     * to the builder.
     *
     * This operation corresponds to appending |binCount| central bins of
     * constant range width
     * W = ((|nextMaxBinBoundary| - |this.maxBinBoundary|) / |binCount|)
     * with the following ranges:
     *
     *   [|this.maxMinBoundary|, |this.maxMinBoundary| + W]
     *   [|this.maxMinBoundary| + W, |this.maxMinBoundary| + 2W]
     *   [|this.maxMinBoundary| + 2W, |this.maxMinBoundary| + 3W]
     *   ...
     *   [|this.maxMinBoundary| + (|binCount| - 2) * W,
     *    |this.maxMinBoundary| + (|binCount| - 2) * W]
     *   [|this.maxMinBoundary| + (|binCount| - 1) * W,
     *    |nextMaxBinBoundary|]
     *
     * @param {number} nextBinBoundary The last added bin boundary (must be
     *     greater than |this.maxMinBoundary|).
     * @param {number} binCount Number of bins to be added (must be positive).
     */
    addLinearBins: function(nextMaxBinBoundary, binCount) {
      if (binCount <= 0)
        throw new Error('Bin count must be positive');

      var curMaxBinBoundary = this.maxBinBoundary;
      if (curMaxBinBoundary >= nextMaxBinBoundary) {
        throw new Error('The last added max boundary must be greater than ' +
            'the current max boundary boundary');
      }

      var binWidth = (nextMaxBinBoundary - curMaxBinBoundary) / binCount;
      for (var i = 1; i < binCount; i++)
        this.addBinBoundary(curMaxBinBoundary + i * binWidth);
      this.addBinBoundary(nextMaxBinBoundary);

      return this;
    },

    /**
     * Add |binCount| exponentially scaled bin boundaries up to
     * |nextMaxBinBoundary| to the builder.
     *
     * This operation corresponds to appending |binCount| central bins with
     * a constant difference between the logarithms of their range min and max
     * D = ((ln(|nextMaxBinBoundary|) - ln(|this.maxBinBoundary|)) / |binCount|)
     * with the following ranges:
     *
     *   [|this.maxMinBoundary|, |this.maxMinBoundary| * exp(D)]
     *   [|this.maxMinBoundary| * exp(D), |this.maxMinBoundary| * exp(2D)]
     *   [|this.maxMinBoundary| * exp(2D), |this.maxMinBoundary| * exp(3D)]
     *   ...
     *   [|this.maxMinBoundary| * exp((|binCount| - 2) * D),
     *    |this.maxMinBoundary| * exp((|binCount| - 2) * D)]
     *   [|this.maxMinBoundary| * exp((|binCount| - 1) * D),
     *    |nextMaxBinBoundary|]
     *
     * This method requires that the current max bin boundary is positive.
     *
     * @param {number} nextBinBoundary The last added bin boundary (must be
     *     greater than |this.maxMinBoundary|).
     * @param {number} binCount Number of bins to be added (must be positive).
     */
    addExponentialBins: function(nextMaxBinBoundary, binCount) {
      if (binCount <= 0)
        throw new Error('Bin count must be positive');

      var curMaxBinBoundary = this.maxBinBoundary;
      if (curMaxBinBoundary <= 0)
        throw new Error('Current max bin boundary must be positive');
      if (curMaxBinBoundary >= nextMaxBinBoundary) {
        throw new Error('The last added max boundary must be greater than ' +
            'the current max boundary boundary');
      }

      var binExponentWidth =
          Math.log(nextMaxBinBoundary / curMaxBinBoundary) / binCount;
      for (var i = 1; i < binCount; i++) {
        this.addBinBoundary(
            curMaxBinBoundary * Math.exp(i * binExponentWidth));
      }
      this.addBinBoundary(nextMaxBinBoundary);

      return this;
    },

    /**
     * Build a tr.v.Numeric from the list of bin boundaries.
     *
     * As explained earlier, this method can be called arbitrarily many times
     * to produce arbitrarily many distinct numerics.
     */
    build: function() {
      var binInfo = {
        underflowBin: new NumericBin(undefined,
            Range.fromExplicitRange(-Number.MAX_VALUE, this.minBinBoundary)),
        overflowBin: new NumericBin(undefined,
            Range.fromExplicitRange(this.maxBinBoundary, Number.MAX_VALUE)),
        centralBins: new Array(this.boundaries_.length - 1)
      };
      for (var i = 0; i < this.boundaries_.length - 1; i++) {
        binInfo.centralBins[i] = new NumericBin(undefined,
            Range.fromExplicitRange(
                this.boundaries_[i], this.boundaries_[i + 1]));
      }

      var numeric = new Numeric(
          this.unit_,
          Range.fromExplicitRange(this.minBinBoundary, this.maxBinBoundary),
          binInfo);
      numeric.allBins.forEach(function(bin) {
        bin.parentNumeric = numeric;
      });
      return numeric;
    }
  };

  /**
   * Create a linearly scaled tr.v.NumericBuilder with |numBins| bins ranging
   * from |range.min| to |range.max|.
   */
  NumericBuilder.createLinear = function(unit, range, numBins) {
    if (range.isEmpty)
      throw new Error('Range must be non-empty');
    return new NumericBuilder(unit, range.min).addLinearBins(
        range.max, numBins);
  };

  function ScalarNumeric(unit, value) {
    if (!(typeof(value) == 'number'))
      throw new Error('Expected value to be number');

    NumericBase.call(this, unit);
    this.value = value;
  }

  ScalarNumeric.prototype = {
    __proto__: NumericBase.prototype,

    asDictInto_: function(d) {
      d.type = 'scalar';
      d.value = this.value;
    },

    toString: function() {
      return this.unit.format(this.value);
    }
  };

  ScalarNumeric.fromDict = function(d) {
    return new ScalarNumeric(tr.v.Unit.fromJSON(d.unit), d.value);
  };

  return {
    NumericBase: NumericBase,
    NumericBin: NumericBin,
    Numeric: Numeric,
    NumericBuilder: NumericBuilder,
    ScalarNumeric: ScalarNumeric
  };
});
